import importlib
import importlib.machinery
import importlib.util
from inspect import stack
from logging import Logger, getLogger
from types import ModuleType

from numpy import where, nan
from pandas import DataFrame, to_datetime, to_numeric
from pandas.core.dtypes.common import is_datetime64_any_dtype

from com_enovation.toolbox.data_handler.data_modifier import data_modification_feature__add_eval_column__lambdas


class DataModificationFeatureFactory:
    """
    TODO: comment

    IDEAS:
    - replace value by
    - replace nan by
    """

    # Instance properties (decorated with @property)
    @property
    def dict_feature_getters(self) -> dict[str, callable]:
        return self._dict_feature_getters

    # Constants
    const__type_columns__str: str = "str"
    const__type_columns__datetime: str = "datetime"
    const__type_columns__date: str = "date"
    const__type_columns__float: str = "float"
    const__type_columns__int: str = "int"

    const__feature_getter_function_prefix: str = "get_data_modification_feature__"

    # Class properties
    _logger: Logger = getLogger(__name__)

    def __init__(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We initialize the catalog of feature getter by introspecting the class and searching for all function
        # that starts with "get_data_modification_feature__"
        self._dict_feature_getters: dict[str, callable] = {
            i_fn_label[len(self.const__feature_getter_function_prefix):]:
                getattr(DataModificationFeatureFactory, i_fn_label)
            for i_fn_label in dir(DataModificationFeatureFactory)
            if (
                    callable(getattr(DataModificationFeatureFactory, i_fn_label))
                    and i_fn_label.startswith(self.const__feature_getter_function_prefix)
            )
        }
        self._logger.debug(f"DataModificationFeatureFactory initialized with '{len(self.dict_feature_getters)}' "
                           f"feature getters: '{', '.join(self.dict_feature_getters)}'.")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def get_modification_feature(
            self,
            str_feature_label: str,
            **kwargs
    ) -> callable:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if str_feature_label not in self.dict_feature_getters:
            self._logger.debug(f"There is not getter for the feature labelled '{str_feature_label}'. Only the "
                               f"following are available: '{', '.join(self.dict_feature_getters)}")
            return None

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return self.dict_feature_getters[str_feature_label](self, **kwargs)

    def get_modifications_features_labels(self) -> list[str]:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _lst_the_return: list[str] = list(self.dict_feature_getters.keys())

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return _lst_the_return

    def get_data_modification_feature__type_columns(
            self,
            dict_columns: dict[str, str],
            b_copy: bool = False
    ):
        """
        Illustration:
        - Sample 1:
            With dict_column = {
                "Jsg - ID": "str",
                "Jsg - Snapshot": "str"
            }
            Will type
            - Column labelled "Jsg - ID" into "string"
            - Column labelled "Jsg - Snapshot" into "string"

        :param dict_columns: type by column
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__type_columns(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            for str_col, str_type in dict_columns.items():

                if str_col in df_data_extract.columns:

                    if str_type == self.const__type_columns__str:

                        # df_the_return[str_col].replace(numpy.nan, "", inplace=True)
                        try:
                            df_the_return[str_col] = df_the_return[str_col].astype(str)
                        except Exception as an_exception:
                            raise Exception(f"Exception when typing column '{str_col}' to '{str_type}'.") \
                                from an_exception

                    elif str_type == self.const__type_columns__datetime:
                        try:
                            df_the_return[str_col] = to_datetime(
                                df_the_return[str_col],
                                # As of 3-Sep-2023: UserWarning: The argument 'infer_datetime_format' is deprecated
                                # infer_datetime_format=True
                            )  # As of 3-Sep-2023: 'normalize' would override the time! .dt.normalize()
                        except Exception as an_exception:
                            raise Exception(f"Exception when typing column '{str_col}' to '{str_type}'.") \
                                from an_exception

                    elif str_type == self.const__type_columns__date:
                        try:
                            df_the_return[str_col] = to_datetime(
                                df_the_return[str_col],
                                # As of 3-Sep-2023: UserWarning: The argument 'infer_datetime_format' is deprecated
                                # infer_datetime_format=True
                            ).dt.date
                        except Exception as an_exception:
                            raise Exception(f"Exception when typing column '{str_col}' to '{str_type}'.") \
                                from an_exception

                        # The below happens when column is empty, and fully made of NaT
                        if str(df_the_return[str_col].dtype).startswith("datetime"):
                            df_the_return[str_col] = df_the_return[str_col].astype('object')

                    elif str_type == self.const__type_columns__float:
                        try:
                            df_the_return[str_col] = to_numeric(
                                arg=df_the_return[str_col],
                                # errors="coerce",
                                downcast="float"
                            )  # .fillna(0)
                        except Exception as an_exception:
                            raise Exception(f"Exception when typing column '{str_col}' to '{str_type}'.") \
                                from an_exception

                    elif str_type == self.const__type_columns__int:
                        try:
                            df_the_return[str_col] = to_numeric(
                                arg=df_the_return[str_col],
                                # errors="coerce",
                                downcast="integer"
                            )  # .fillna(0)
                        except Exception as an_exception:
                            raise Exception(f"Exception when typing column '{str_col}' to '{str_type}'.") \
                                from an_exception
                    else:
                        raise Exception(
                            f"When transcoding the date extract, we did not know what to do with column '{str_col}' "
                            f"as we don't know the target type '{str_type}'...")

                else:
                    self._logger.warning(f"Column '{str_col}' is to be typed as '{str_type}', but it does not exist "
                                         f"in the data extract.")

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__type_columns

    def get_data_modification_feature__replace_values(
            self,
            dict_columns: dict[str, dict],
            b_copy: bool = False
    ):
        """
        Illustration:
        - Sample 1:
            With dict_column = {
                "Jsg - ID": {
                    "-": 0,
                    "True": 1
                    "False": 0
                }
            }
            Will replace
            - In column labelled "Jsg - ID":
                - Value "-" into 0
                - Value "True" into 1
                - Value "False" into 0

        :param dict_columns: by column, the values to be replaced
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__replace_values(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            # We loop through each and every column to process
            for str_col, dict_transco in dict_columns.items():

                # If the column exist in the dataframe
                if str_col in df_data_extract.columns:

                    # We loop through each and every transco
                    for k_from, v_to in dict_transco.items():
                        df_the_return[str_col].replace(k_from, v_to, inplace=True)

                else:
                    self._logger.warning(f"Column '{str_col}' is to be modified, with values to be transcoded, but it "
                                         f"does not exist in the data extract.")

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__replace_values

    def get_data_modification_feature__replace_missing_values(
            self,
            dict_columns: dict,
            b_copy: bool = False
    ):
        """
        Illustration:
        - Sample 1:
            With dict_column = {
                "Jsg - ID": 0,
                "Jsg - Snapshot": "one"
            }
            Will replace missing values (aka NaN)
            - By 0 in column labelled "Jsg - ID"
            - By "one" in column labelled "Jsg - Snapshot"

        :param dict_columns: by column, the value to replace missing values
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__replace_missing_values(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            # We loop through each and every column to process
            for str_col, v_value in dict_columns.items():

                # If the column exist in the dataframe
                if str_col in df_data_extract.columns:

                    df_the_return[str_col].fillna(value=v_value, inplace=True)

                else:
                    self._logger.warning(f"Column '{str_col}' is to be modified, with missing values to be set to "
                                         f"'{v_value}', but it does not exist in the data extract.")

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__replace_missing_values

    def get_data_modification_feature__add_eval_columns(
            self,
            str_expressions: list[str] | str,
            local_dict: dict = None,
            lst_dependencies: list[str] = None,
            b_copy: bool = False
    ):
        """
        Illustration:
        - Sample 1:
            With str_expressions=f" " "
                DQ_01=StatusA+StatusB=={[
                    "OpenOpen",
                    "OpenInactive",
                    "InactiveOpen",
                    "InactiveInactive",
                    "ClosedClosed"
                ]}
                DQ_02=StatusA=='Open'
                " " "
            We add 2 columns:
                - DQ_01 equal to True in case concatenation Status A + B is among the ones in the list (authorized)
                - DQ_02 equal to True if Status A is "Open"
        - Sample 2:
            With str_expression = "FullName=`Last Name` + @df.assign(Tmp=', ').Tmp + `First Name`
            We are adding one column "Guillard, Jean-Sébastien"

        :param lst_dependencies:
        :param local_dict:
        :param str_expressions: the expression to add new columns through eval
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if isinstance(str_expressions, list):
            str_expressions = "\n".join(str_expressions)

        # We enrich the dictionary of lambdas
        if local_dict is None:
            local_dict = {}

        local_dict.update(
            {
                i_fn_label[1:]: getattr(data_modification_feature__add_eval_column__lambdas, i_fn_label)
                for i_fn_label in dir(data_modification_feature__add_eval_column__lambdas)
                if (
                    callable(getattr(data_modification_feature__add_eval_column__lambdas, i_fn_label))
                    and i_fn_label.startswith("_")
                )
            }
        )

        if lst_dependencies is None:
            lst_dependencies = []

        for i_str_dependency in lst_dependencies:
            if "." not in i_str_dependency:
                raise Exception(
                    f"The dependency '{i_str_dependency}' does not contain any '.', so it cannot be imported... "
                    f"Dependencies should be in the form 'name.of.a.module.followed.by.function_name")

            else:
                i_lst_dependency: list[str] = i_str_dependency.rsplit(".", 1)

                i_str_module_name: str = i_lst_dependency[0]
                i_str_function_name: str = i_lst_dependency[1]

                i_mdl_module: ModuleType = importlib.import_module(i_str_module_name)
                i_fn_function: callable = getattr(i_mdl_module, i_str_function_name)

                local_dict[i_str_function_name] = i_fn_function

        def fn__add_eval_column(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            local_dict["df"]=df_the_return

            df_the_return.eval(str_expressions, inplace=True, local_dict=local_dict)

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__add_eval_column

    def get_data_modification_feature__filter(
            self,
            str_expressions: str,
            b_copy: bool = False
    ):
        """
        Illustration:
        - Sample 1:
            With str_expressions=f" " "
                StatusA+StatusB=={[
                    "OpenOpen",
                    "OpenInactive",
                    "InactiveOpen",
                    "InactiveInactive",
                    "ClosedClosed"
                ]}
                " " "
            We filter rows where :
                - DQ_01 equal to True in case concatenation Status A + B is among the ones in the list (authorized)
                - DQ_02 equal to True if Status A is "Open"

        :param str_expressions: the expression to add new columns through eval
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__filter(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            df_the_return.query(str_expressions, inplace=True)

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__filter

    def get_data_modification_feature__drop_columns(
            self,
            lst_columns_to_drop: list[str],
            b_copy: bool = False
    ):
        """
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__drop_columns(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We check the columns to drop are effectively in the dataframe
            _lst_the_missing_columns: list[str] = [
                i_col
                for i_col in lst_columns_to_drop
                if i_col not in df_data_extract.columns
            ]
            if len(_lst_the_missing_columns) > 0:
                self._logger.error(
                    f"Function '{stack()[0].filename} - {stack()[0].function}' - "
                    f"The following columns are to be dropped, but they could not be found in the dataframe:"
                    f"\n\t- missing columns: {', '.join(_lst_the_missing_columns)}"
                    f"\n\t- dataframe columns: {', '.join(df_data_extract.columns)}")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            df_the_return.drop(
                columns=[
                    i_col
                    for i_col in lst_columns_to_drop
                    if i_col in df_data_extract.columns
                ],
                inplace=True
            )

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__drop_columns

    def get_data_modification_feature__audit_column__changed(
            self,
            s_column_to_audit: str,
            s_column_audit: str,
            lst_key_columns: list[str] = None,
            b_first_as_changed: bool = False,
            b_copy: bool = False
    ):
        """
        Function to add an audit column that tracks whether the audited column changed from one row to the next.
        In case the parameter lst_key_columns is set, the dataframe is grouped accordingly before tracking changes
        across rows.

        The column audit is "True" if value changed, "False" otherwise.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__audit_column__changed(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We check the column to audit is effectively in the dataframe
            if s_column_to_audit not in df_data_extract.columns:
                raise Exception(f"The column to audit '{s_column_to_audit}' does not exist in the dataframe, which "
                                f"contains the following columns: '{', '.join(df_data_extract.columns)}'.")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            if lst_key_columns:

                # We list the key columns which effectively exist in the dataframe
                _lst_the_key_columns: list[str] = [
                    i_col
                    for i_col in lst_key_columns
                    if i_col in df_data_extract.columns
                ]

                # In case we are missing some key columns, we log an error
                if len(_lst_the_key_columns) != len(lst_key_columns):
                    _lst_the_missing_col: list[str] = [
                        i_col
                        for i_col in lst_key_columns
                        if i_col not in df_data_extract.columns
                    ]

                    self._logger.error(
                        f"Function '{stack()[0].filename} - {stack()[0].function}' - "
                        f"The following columns, to group by the dataframe, do not exist. They will be ignored..."
                        f"\n\t- missing columns: {', '.join(_lst_the_missing_col)}"
                        f"\n\t- dataframe columns: {', '.join(df_data_extract.columns)}")

                if b_first_as_changed:
                    df_the_return[s_column_audit] = (
                                                        df_the_return[s_column_to_audit].ne(
                                                            df_the_return.groupby(by=_lst_the_key_columns)[
                                                                s_column_to_audit].shift())
                                                    ) | (
                                                        df_the_return.groupby(by=_lst_the_key_columns)[
                                                            s_column_to_audit].shift().isna()
                                                    )
                else:
                    df_the_return[s_column_audit] = (
                                                        df_the_return[s_column_to_audit].ne(
                                                            df_the_return.groupby(by=_lst_the_key_columns)[
                                                                s_column_to_audit].shift())
                                                    ) & (
                                                        ~df_the_return.groupby(by=_lst_the_key_columns)[
                                                            s_column_to_audit].shift().isna()
                                                    )

            else:
                raise Exception(f"Not implemented yet.")

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__audit_column__changed

    def get_data_modification_feature__audit_column__stamped(
            self,
            s_column_to_audit: str,
            s_column_audit: str,
            s_column_to_stamp: str,
            lst_key_columns: list[str] = None,
            b_copy: bool = False
    ):
        """
        Function to add an audit column that stamp rows with the audited column not changing using the value in the
        "column to stamp" at the time the audited value was set first.
        In case the parameter lst_key_columns is set, the dataframe is grouped accordingly before tracking changes
        across rows.

        The column audit is set to the value from the "column to stamp" when the audited value changed.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__audit_column__stamped(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We check the column to audit is effectively in the dataframe
            if s_column_to_audit not in df_data_extract.columns:
                raise Exception(f"The column to audit '{s_column_to_audit}' does not exist in the dataframe, which "
                                f"contains the following columns: '{', '.join(df_data_extract.columns)}'.")
            if s_column_to_stamp not in df_data_extract.columns:
                raise Exception(f"The column to stamp '{s_column_to_stamp}' does not exist in the dataframe, which "
                                f"contains the following columns: '{', '.join(df_data_extract.columns)}'.")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            if lst_key_columns is None:
                _lst_the_key_columns: list[str] = []

            else:

                # We list the key columns which effectively exist in the dataframe
                _lst_the_key_columns: list[str] = [
                    i_col
                    for i_col in lst_key_columns
                    if i_col in df_data_extract.columns
                ]

                # In case we are missing some key columns, we log an error
                if len(_lst_the_key_columns) != len(lst_key_columns):
                    _lst_the_missing_col: list[str] = [
                        i_col
                        for i_col in lst_key_columns
                        if i_col not in df_data_extract.columns
                    ]

                    self._logger.error(
                        f"Function '{stack()[0].filename} - {stack()[0].function}' - "
                        f"The following columns, to group by the dataframe, do not exist. They will be ignored..."
                        f"\n\t- missing columns: {', '.join(_lst_the_missing_col)}"
                        f"\n\t- dataframe columns: {', '.join(df_data_extract.columns)}")

            # First, we audit the column to spot when value changed: we get true/ false in the audit column
            df_the_return = self.get_data_modification_feature__audit_column__changed(
                s_column_to_audit=s_column_to_audit,
                s_column_audit=s_column_audit,
                lst_key_columns=_lst_the_key_columns,
                b_first_as_changed=True,
                b_copy=False
            )(df_the_return)

            # In the where clause below, if column x 's_column_to_stamp' is of type datetime64[ns], an exception is
            # thrown
            if is_datetime64_any_dtype(df_the_return[s_column_to_stamp].dtype):
                _x = df_the_return[s_column_to_stamp].dt.date
                self._logger.warning(f"We have to transcode column '{s_column_to_stamp}' from "
                                     f"'{df_the_return[s_column_to_stamp].dtype}' to '{_x.dtype}'")
            else:
                _x = df_the_return[s_column_to_stamp]

            # Whenever the audit changed is True, we set the stamp value in the audit column
            df_the_return[s_column_audit] = where(
                df_the_return[s_column_audit],
                _x,
                nan
            )

            df_the_return[s_column_audit] =\
                df_the_return.groupby(by=_lst_the_key_columns)[s_column_audit].ffill()

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__audit_column__stamped

    def get_data_modification_feature__sort(
            self,
            by: str | list[str],
            axis: int = 0,
            ascending: bool | list[bool] = True,
            inplace: bool = False,
            kind: str = "quicksort",
            na_position: str = "last",
            ignore_index: bool = False,
            b_copy: bool = False
    ):
        """
        Function to sort the dataframe. Parameters are documented in the official pandas.DataFrame.sort_values.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__sort(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We check the columns to sort is effectively in the dataframe
            if isinstance(by, str):
                _lst_by: list[str] = [by]
            elif isinstance(by, list):
                _lst_by: list[str] = by
            else:
                raise Exception(f"Parameter by is of type '{type(by)}', which is not expected.")

            _lst_missing_by: list[str] = [i_col for i_col in _lst_by if i_col not in df_data_extract.columns]
            if len(_lst_missing_by) > 0:
                self._logger.error(f"Columns to sort dataframe by are missing: '{', '.join(_lst_missing_by)}'.")
                _lst_by = [i_col for i_col in _lst_by if i_col not in _lst_missing_by]
                if len(_lst_by) == 0:
                    raise Exception(f"Among the columns to sort, none exist in the DataFrame, which is not expected.")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            df_the_return = df_the_return.sort_values(
                by=by,
                axis=axis,
                inplace=inplace,
                ascending=ascending,
                kind=kind,
                na_position=na_position,
                ignore_index=ignore_index
            )

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__sort

    def get_data_modification_feature__repeat_last(
            self,
            s_column_to_repeat: str,
            s_column_repeat: str,
            lst_key_columns: list[str] = None,
            b_copy: bool = False
    ):
        """
        Function that add a column that repeats the last value.

        Eg:
            - Function called with parameters:
                - s_column_to_repeat: Measure
                - s_column_repeat: Last Measure
                - lst_key_columns: Key
            - test data and results below
                |-------------------------------------|   |---------------|
                | Key     | Date          | Measure   |   | Last Measure  |
                |-------------------------------------|   |---------------|
                | JSG     | 01-Jan-2023   | 44 years  |   | 44 years      |
                | BCH     | 01-Jan-2023   | 44 years  |   | 44 years      |
                | JSG     | 21-May-2023   | 45 years  |   | 44 years      |
                | BCH     | 21-May-2023   | 44 years  |   | 44 years      |
                | JSG     | 24-Jul-2023   | 45 years  |   | 45 years      |
                | BCH     | 24-Jul-2023   | 45 years  |   | 44 years      |
                | JSG     | 10-Nov-2023   | 45 years  |   | 45 years      |
                | BCH     | 10-Nov-2023   | 45 years  |   | 45 years      |
                | JSG     | 31-Dec-2023   | 45 years  |   | 45 years      |
                | BCH     | 31-Dec-2023   | 45 years  |   | 45 years      |
                |-------------------------------------|   |---------------|
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__repeat_last(
                df_data_extract: DataFrame,
        ) -> DataFrame:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We check the column to repeat is effectively in the dataframe
            if s_column_to_repeat not in df_data_extract.columns:
                raise Exception(f"The column to repeat '{s_column_to_repeat}' does not exist in the dataframe, which "
                                f"contains the following columns: '{', '.join(df_data_extract.columns)}'.")

            # We check the repeat column is not in the dataframe
            if s_column_repeat in df_data_extract.columns:
                raise Exception(f"The repeat column '{s_column_repeat}' does exist in the dataframe, which is not "
                                f"expected.")

            if b_copy:
                df_the_return: DataFrame = df_data_extract.copy()
            else:
                df_the_return: DataFrame = df_data_extract

            if lst_key_columns is None:
                _lst_the_key_columns: list[str] = []

            else:

                # We list the key columns which effectively exist in the dataframe
                _lst_the_key_columns: list[str] = [
                    i_col
                    for i_col in lst_key_columns
                    if i_col in df_data_extract.columns
                ]

                # In case we are missing some key columns, we log an error
                if len(_lst_the_key_columns) != len(lst_key_columns):
                    _lst_the_missing_col: list[str] = [
                        i_col
                        for i_col in lst_key_columns
                        if i_col not in df_data_extract.columns
                    ]

                    self._logger.error(
                        f"Function '{stack()[0].filename} - {stack()[0].function}' - "
                        f"The following columns, to group by the dataframe, do not exist. They will be ignored..."
                        f"\n\t- missing columns: {', '.join(_lst_the_missing_col)}"
                        f"\n\t- dataframe columns: {', '.join(df_data_extract.columns)}")

            # First, we audit the column to spot when value changed: we get true/ false in the audit column
            df_the_return = self.get_data_modification_feature__audit_column__changed(
                s_column_to_audit=s_column_to_repeat,
                s_column_audit=s_column_repeat,
                lst_key_columns=_lst_the_key_columns,
                b_first_as_changed=False,
                b_copy=False
            )(df_the_return)

            # Whenever the audit changed is False, we set the repeated value as the current value
            df_the_return[s_column_repeat] = where(
                df_the_return[s_column_repeat],
                nan,
                df_the_return[s_column_to_repeat]
            )

            df_the_return[s_column_repeat] =\
                df_the_return.groupby(by=_lst_the_key_columns)[s_column_repeat].ffill()

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return df_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__repeat_last



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_AuditColumnChanged(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_audit_column_changed(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__),
                              '01.data_extract_for_audit_column.xlsx'))
        )

        # 1. We audit the column, grouping on one single column, first occurrence returns False
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__changed(
            s_column_to_audit="Jsg - Audited Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID", "a column which does not exist"]
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                True,

                True
            ],
            msg="1. We audit the column, grouping on one single column, first occurrence returns False"
        )

        # 2. We audit the column, grouping on one single column, first occurrence returns True
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__changed(
            s_column_to_audit="Jsg - Audited Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID"],
            b_first_as_changed=True
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                True,
                False,
                True,
                True,
                False,
                True,
                True,
                True,
                False,
                True,
                True,
                False,
                True,
                True,
                True,

                True
            ],
            msg="2. We audit the column, grouping on one single column, first occurrence returns True"
        )

        # 3. We audit the column, grouping on two columns, first occurrence returns False
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__changed(
            s_column_to_audit="Jsg - Audited Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID", "Jsg - Snapshot"],
            b_first_as_changed=False
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                True,

                True
            ],
            msg="3. We audit the column, grouping on two columns, first occurrence returns False"
        )

        # 4. We audit the column, grouping on two columns, first occurrence returns True
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__changed(
            s_column_to_audit="Jsg - Audited Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID", "Jsg - Snapshot"],
            b_first_as_changed=True
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                True,
                False,
                True,
                True,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                True,
                True,
                True,

                True
            ],
            msg="4. We audit the column, grouping on two columns, first occurrence returns True"
        )

        # 5. Exception if we audit a column which does not exist
        with self.assertRaisesRegex(
                Exception,
                f"The column to audit 'Jsg - not existing column' does not exist in the dataframe, which contains the "
                f"following columns: 'Jsg - ID, Jsg - Snapshot, Jsg - Audited Column'."
        ):
            DataModificationFeatureFactory(). \
                get_data_modification_feature__audit_column__changed(
                s_column_to_audit="Jsg - not existing column",
                s_column_audit="Jsg - Audited Column_audit",
                lst_key_columns=["Jsg - ID"]
            )(df_data_extract[["Jsg - ID", "Jsg - Snapshot", "Jsg - Audited Column"]])

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path
from unittest import TestCase

from com_enovation.toolbox.excel_dashboard.excel_dashboarder import ExcelDashboarder


class TestExcelDashboarder(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC01.excel.xlsm')),
            dict_config={
                "vba": {
                    "workbook.add_vba_project": os.path.join(os.path.dirname(__file__), 'vbaProject.bin'),
                    # "workbook.set_vba_name": "jsg_wb",
                    # "worksheet.set_vba_name": {
                    #     "jsg_sheet_1": "Sheet1",
                    #     "jsg_sheet_2": "Sheet2"
                    # }
                },
                "sheets": {
                    "jsg_sheet_1": {
                        "widgets": {
                            "test_first_label": {
                                "address": "B2",
                                "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                "config": {
                                    "narrative": "the first text is '{0}' and the second is '{1}'."
                                },
                            }
                        }
                    },
                    "jsg_sheet_2": {
                        "widgets": {
                            "test_second_label": {
                                "address": "C3",
                                "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                "config": {
                                    "narrative": "the first text is '{0}' and the second is '{1}'."
                                },
                            }
                        }
                    }
                }
            },
            **{"tokens": ["ici c'est paris", "mais pourquoi donc?"]}
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returned")



import json
import unittest

import click
from click.testing import CliRunner, Result
from logging import Logger, getLogger
from inspect import stack

from com_enovation.helper.click.python_literal_argument_and_option import PythonLiteralOption


@click.command('my-test')
@click.option('--a-list', cls=PythonLiteralOption, type=list, default=[])
@click.option('--a-dict', cls=PythonLiteralOption, type=dict, default={})
def my_test(a_list, a_dict):

    # 1. We check the list
    if isinstance(a_list, list):
        if a_list != ["o11", "o12", "o13"]:
            raise Exception(f"Parameter a_list is not as expected, but is '{a_list}'.")
    else:
        raise Exception(f"Option a_list is not a list, but a '{type(a_list)}'.")

    # We check the dictionary
    if isinstance(a_dict, dict):
        dict_expected: dict = {"k1": "v1", "k2": "v2"}
        if len(a_dict) != len(dict_expected):
            raise Exception(f"Option a_dict has '{len(a_dict)}' records, while '{len(dict_expected)}' is expected.")

        dict_diff: dict = {
            key: value for (key, value) in a_dict.items() if (
                (key not in dict_expected.keys()) | (dict_expected[key] != value)
            )
        }
        if len(dict_diff) > 0:
            raise Exception(f"Option a_dict is not as expected, but has the following items in difference: "
                            f"'{json.dumps(dict_diff)}'")
    else:
        raise Exception(f"Option a_dict is not a dict, but a '{type(a_dict)}'.")


class TestPythonLiteralOption(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_date_predictability(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_runner = CliRunner()

        the_result: Result = the_runner.invoke(
            cli=my_test,
            args=[
                '--a-list', '["o11", "o12", "o13"]',
                '--a-dict', '{"k1": "v1", "k2": "v2"}',
            ],
            catch_exceptions=False,
        )

        self.assertEqual(
            0,
            the_result.exit_code,
            f"Unexpected exit_code, with stdout: {the_result.stdout}"
        )



from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame

from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com_enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class TestFunctionResampleMeasures(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_vanilla(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_df_dataset: DataFrame = DataFrame(
            columns=[
                DatePredictabilityComputer.str__input__column_label__key,
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
            ],
            data=[
                ["abc", "2021-4-26", "2021-5-22"],
                ["abc", "2021-4-27", "2021-6-5"],
                ["abc", "2021-4-30", "2021-6-3"],
                ["abc", "2021-5-1", "2021-6-4"],
                ["abc", "2021-5-6", "2021-5-29"],
                ["abc", "2021-5-7", "2021-5-25"],
                ["abc", "2021-5-9", "2021-5-21"],
                ["abc", "2021-5-11", "2021-5-19"],
                ["abc", "2021-5-13", "2021-5-16"],
                ["abc", "2021-5-16", "2021-5-17"],
                ["abc", "2021-5-18", "2021-5-21"],
                ["abc", "2021-5-23", "2021-5-14"],
                ["abc", "2021-5-25", "2021-5-27"],
                ["abc", "2021-5-28", "2021-5-27"],
                ["abc", "2021-5-29", "2021-5-20"],
                ["abc", "2021-6-6", "2021-5-19"],

                ["bcd", "2021-6-6", "2021-5-19"],

                ["def", "2019-6-6", "2020-5-19"],
                ["def", "2021-6-6", "2021-5-19"],
            ]
        )

        # We type the columns
        the_df_dataset = PandasDataframeTyper.type(
            df_to_type=the_df_dataset,
            dict_columns_to_type={
                DatePredictabilityComputer.str__input__column_label__date: PandasDataframeTyper.str__type__date,
                DatePredictabilityComputer.str__input__column_label__measure: PandasDataframeTyper.str__type__date,
            },
            b_strict=False
        )

        the_df_resampled_measures: DataFrame = DatePredictabilityComputer()._resample_measures(
            df_measures=the_df_dataset
        )

        self.assertEqual(
            first=(775, 3),
            second=the_df_resampled_measures.shape,
            msg="unexpected shape"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import inspect
from unittest import TestCase
import logging
import os

from com_enovation.helper.excel_loader import ExcelLoader, ExcelLoaderBean


class TestXlsFileHelper(TestCase):

    _logger: logging.Logger = logging.getLogger(__name__)

    def test_function_read_xls_as_dataframe(self):
        """
        Test the function to read_xls_as_dataframe:
            - Test case 1: the xls file exists, and is correctly loaded
            - Test case 2: the xls file does not exist, and an error is thrown
            - Test case 3: the xls file exists, but is corrupted, and an error is thrown
        """
        self._logger.debug(f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' "
                           f"is called.")

        the_loader: ExcelLoader = ExcelLoader()

        # Test case 1
        the_test_case_1_bean: ExcelLoaderBean = the_loader.load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC01.ExcelFileWithTables.xlsx')
        )

        self._logger.debug("Test case 1 executed")

        self._logger.debug(f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' "
                           f"is returning.")



import os
from inspect import stack
from logging import getLogger, Logger
from pathlib import Path
from unittest import TestCase

import xlsxwriter
from openpyxl import load_workbook

from com_enovation.toolbox.excel_dashboard.excel_dashboarder import ExcelDashboarder
from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker
from com_enovation.toolbox.excel_dashboard.widgets.cover_sheet.widget_cover_sheet import WidgetCoverSheet


class TestWidget(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_config_checker(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker: ConfigChecker = ConfigChecker(
            lst_widgets_json_schemas=[WidgetCoverSheet.get_dict_config_schema()],
            str_base_json_schema_id=WidgetCoverSheet.get_widget_id()
        )

        _the_config_checker.validate(
            dict_to_validate={"narrative": "yo, this is fantastic {0} et pourquoi pas {1}"},
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unitary(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We instantiate the widget
        _the_widget: WidgetCoverSheet = WidgetCoverSheet(
            str_address="B2",
            dict_parameters_map={"tokens": "jsg_tokens"},
            dict_config={"narrative": "yo, this is fantastic {0} et pourquoi pas {1}"},
            dict_raw_parameters={"jsg_tokens": ["ici c'est paris", "mais pourquoi donc?"]},
            str_sheet_name="abc",
            dict_default_format={}
        )

        # We print it into an excel spreadsheet
        _the_workbook = xlsxwriter.Workbook(Path(os.path.join(os.path.dirname(__file__), 'TC02.excel.xlsx')))
        _the_worksheet = _the_workbook.add_worksheet()

        # We print the widget
        _the_widget.write_to_excel(
            wb_workbook=_the_workbook,
            ws_worksheet=_the_worksheet
        )

        # We close the workbook
        _the_workbook.close()

        # We now read the excel spreadsheet, and assert the result is as expected

        _the_workbook = load_workbook(
            filename=Path(os.path.join(os.path.dirname(__file__), 'TC02.excel.xlsx')),
            data_only=True
        )
        _the_worksheet = _the_workbook.active
        _the_string = _the_worksheet["B2"].value

        self.assertEqual(
            first=_the_string,
            second=_the_widget.str_narrative.format(*_the_widget.lst_tokens),
            msg="The message stored in the excel spreadsheet is not as expected..."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_application(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC03.excel.xlsx')),
            dict_config={
                "sheets": {
                    "test_sheet_name": {
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "C3",
                                "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                "config": {
                                    "narrative": "the first text is '{0}' and the second is '{1}'."
                                },
                            }
                        }
                    }
                }
            },
            **{"tokens": ["ici c'est paris", "mais pourquoi donc?"]}
        )

        # We now read the excel spreadsheet, and assert the result is as expected

        _the_workbook = load_workbook(
            filename=Path(os.path.join(os.path.dirname(__file__), 'TC03.excel.xlsx')),
            data_only=True
        )
        _the_worksheet = _the_workbook.active
        _the_string = _the_worksheet["C3"].value

        self.assertEqual(
            first=_the_string,
            second="the first text is '{0}' and the second is '{1}'.".format(
                *["ici c'est paris", "mais pourquoi donc?"]
            ),
            msg="The message stored in the excel spreadsheet is not as expected..."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returned")



import os
import signal
import unittest
from time import sleep
# import json

from click.testing import CliRunner
from logging import Logger, getLogger
from inspect import stack
from com_enovation.enov import enov


class TestDatePredictability(unittest.TestCase):

    import logging.config

    logging.basicConfig(
        format='%(levelname)s [%(asctime)s]: %(message)s',
        level=logging.DEBUG
    )

    logging.info('com.enovation: tests logging initialized')

    _logger: Logger = getLogger(__name__)

    def test_bubbles(self):
        """
        Note: When calling function 'CliRunner.invoke', we add a tag "# noinspection PyTypeChecker" not to raise a
        warning due to function "invoke" expecting function that we call to be a "BaseCommand", which does not seem to
        be the case...

        python3 ./src/enov.py --verbose df-load-xls ./tests/test_enov/test_dash/test_bubbles/01_input_file.xlsx df_raw
        dict-set-json "{\"properties\": {\"title\":\"JSG\"},\"sheets\": {\"sheet abc\": {\"properties\": {},
        \"widgets\":[{\"id\":\"https://enovation.com/dash_application/bubbles\",\"properties\":{\"data\":\"df_raw\"}
        }]}}}" config ds-dash config


        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_runner = CliRunner()

        # ###############################################################
        # 3. We launch the ds_dash
        if hasattr(os, 'fork'):
            int_childpid: int = os.fork()

            if int_childpid == 0:
                # noinspection PyTypeChecker
                the_runner.invoke(
                        cli=enov,
                        args=[
                            '--verbose',

                            # We load the Excel file
                            'df-load-xls',
                            os.path.join(os.path.dirname(__file__), '01_input_file.xlsx'),
                            'df_raw',

                            # We set the widgets configuration
                            'dict-set-json',
                            '{'
                            '   "properties": {"title":"JSG"},'
                            '   "sheets": {'
                            '       "sheet abc": {'
                            '           "properties": {},'
                            '           "widgets": ['
                            '               {'
                            '                   "id": "https://enovation.com/dash_application/bubbles",'
                            '                   "properties": {'
                            '                       "data":"df_raw",'
                            '                       "px.scatter":{'
                            '                           "x":"Start Date",'
                            '                           "y":"Probability",'
                            '                           "animation_frame":"Timestamp",'
                            '                           "animation_group":"Project",'
                            '                           "size":"Effort",'
                            '                           "color":"Territory",'
                            '                           "hover_name":"Project",'
                            # '                           "log_x":True,'
                            # '                           "size_max":55,'
                            # '                           "range_x":[100, 100000],'
                            # '                           "range_y":[25, 90]'
                            '                       }'
                            '                   }'
                            '               }'
                            '           ]'
                            '       }'
                            '   }'
                            '}',
                            'config',

                            # We load the predictability bean
                            'ds-dash',
                            'config'
                            # os.path.join(os.path.dirname(__file__), '03_actual_results.xls'),
                            # 'df_predictability',
                            #
                            # # We graph
                            # 'dp-graph',
                            # 'df_predictability'
                        ],
                        catch_exceptions=True
                    )

            else:
                # in the parent
                sleep(60)
                os.kill(int_childpid, signal.SIGINT)

        else:
            self._logger.warning(f"You must be executing on Windows, as os.fork() could be executed...")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        raise Exception(f"Bubbles as div is effectively NOK, as it contains constants 'Sp - xxx' that might not be "
                        f"in the provided data... Bubbles widget is to be revisited!")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestWorksheetAddTable(TestCase):

    _logger: Logger = getLogger(__name__)

    _dict_a_full_format: dict = {

        # to turn on or off the autofilter in the header row. It is on by default
        "autofilter": False,

        # to turn on or off the header row in the table. It is on by default
        "header_row": True,

        # to create columns of alternating color in the table. It is on by default
        "banded_columns": False,

        # to create rows of alternating color in the table. It is on by default
        "banded_rows": True,

        # to highlight the first column of the table. The type of highlighting will depend on the style of the table.
        # It may be bold text or a different color. It is off by default
        "first_column": True,

        # to highlight the last column of the table. The type of highlighting will depend on the style of the table.
        # It may be bold text or a different color. It is off by default
        "last_column": True,

        # to set the style of the table. Standard Excel table format names should be used (with matching capitalization)
        "style": "Table Style Medium 15",

        # to turn on the total row in the last row of a table. It is distinguished from the other rows by a different
        # formatting and also with dropdown SUBTOTAL functions
        "total_row": True,

        # default tables are named Table1, Table2, etc. The name parameter can be used to set the name of the table
        "name": "table_name",

        # to set properties for columns within the table
        "columns": [
            {
                # the column header
                "header": "the header",

                # the format for the column header
                "header_format": {
                    "font_name": "Times New Roman",
                    "font_size": 15,
                },

                # column formulas applied using the column formula property
                "formula": "=today()",

                # total captions are specified via the columns property and the total_string sub properties
                "total_string": "total",

                # total functions are specified via the columns property and the total_function sub properties
                # total_function is a string among: average, count_nums, count, max, min, std_dev, sum, var
                "total_function": "average",

                # to set a calculated value for the total_function using the total_value sub property. Only necessary
                # when creating workbooks for applications that cannot calculate the value of formulas automatically.
                # This is similar to setting the value optional property in write_formula()
                "total_value": 15,

                # "the format for the column"
                "format": {
                    "font_name": "Times New Roman",
                    "font_size": 15,
                }

            }
        ]
    }

    def test_01_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/add_table/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/add_table/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                f"Additional properties are not allowed \\('jsg' was unexpected\\)"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_empty(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/add_table/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate={}
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_04_incorrect_format(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/add_table/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format)
        _the_altered_dict["columns"] = [{
            # the column header
            "header": "the header",

            # the format for the column header
            "header_format": {
                "font_name": "Times New Roman",
                "font_size": 15,
                "jsg": "bou"
            },

            # column formulas applied using the column formula property
            "formula": "=today()",

            # total captions are specified via the columns property and the total_string sub properties
            "total_string": "total",

            # total functions are specified via the columns property and the total_function sub properties
            # total_function is a string among: average, count_nums, count, max, min, std_dev, sum, var
            "total_function": "average",

            # to set a calculated value for the total_function using the total_value sub property. Only necessary
            # when creating workbooks for applications that cannot calculate the value of formulas automatically.
            # This is similar to setting the value optional property in write_formula()
            "total_value": 15,

            # "the format for the column"
            "format": {
                "font_name": "Times New Roman",
                "font_size": 15,
            }

        }]

        with self.assertRaisesRegex(
                ValidationError,
                f"Additional properties are not allowed \\('jsg' was unexpected\\)"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import datetime
import re
from datetime import timedelta

from numpy import nan
from pandas import DataFrame, Series


def _enov_timedelta(
        s_column: Series,
        f_delta_days: float
) -> Series:
    """
    Function that can be called through function "pandas.DataFrame.eval":
    `col target label`=@enov_timedelta(`col source label`, 15)

    Exceptions are thrown: FutureWarning: In a future version, object-dtype columns with all-bool values
    will not be included in reductions with bool_only=True. Explicitly cast to bool dtype instead.

    f_delta_days does not work with negative value as of 9-Apr-2023, as python cannot parse the function call... and
    raise an AttributeError xception with message 'UnaryOp object has no attribute value'.
    """
    return s_column + timedelta(days=f_delta_days)


def _enov_concatenate_boolean_tags(
        s_column: list[Series],
        s_labels: list[str]
):
    """
    TODO: check s_column has at least one series
    TODO: check all series have the same number of records
    TODO: check the s label length is equal to the s column length
    TODO: check the index is equal across columns
    """
    s_the_return: Series = Series([""] * len(s_column[0]), index=s_column[0].index)
    # We replace True/ False values by the labels

    for i_col in range(len(s_column)):
        # s_the_return = s_the_return.str.cat(s_column[i_col].replace(
        #     True, s_labels[i_col]+"|").replace(False, ""))

        s_the_return = s_the_return + s_column[i_col].replace(
            True, "|" + s_labels[i_col]).replace(False, "")

    # In case several tags were concatenated, we just remove the very first '|'
    s_the_return = s_the_return.str.slice(start=1)

    return s_the_return


def _enov_regex(
        s_column: Series,
        str_regex: str
):
    s_the_return: Series = s_column.str.findall(pat=str_regex, flags=re.MULTILINE)
    return s_the_return


def _enov_list_to_dates(
        s_column: Series,
        pattern: str
):
    s_the_return: Series = s_column.apply(
        func=lambda a_list: [datetime.datetime.strptime(date, pattern).date() for date in a_list]
    )
    return s_the_return


def _enov_list_max(
        s_column: Series
):
    s_the_return: Series = s_column.apply(
        func=lambda a_list: max(a_list) if len(a_list) > 0 else nan
    )
    return s_the_return


def _enov_lambda(
        columns: list[Series] | Series,
        s_lambda: str
) -> Series:
    """
    Illustration:
        - Sample 1:
            With str_expressions="Enov_Label=@enov_lambda(columns=['FirstName', 'LastName'], s_lambda='row[\\'LastName\\']+\\', \\'+row[\\'FirstName\\']')"
            We are adding one column "Guillard, Jean-Sébastien"

        :param lst_dependencies:
        :param local_dict:
        :param str_expressions: the expression to add new columns through eval
        :param b_copy: whether dataframe is to be copied before modified
        :return: modified dataframe
    
    """

    if isinstance(columns, Series):
        columns = [columns]

    _df: DataFrame = DataFrame({
        i_col.name: i_col for i_col in columns
    })

    _s_the_return: Series = _df.apply(func=lambda row: eval(s_lambda), axis = 1)
    return _s_the_return


def _enov_get_record_at_index_from_tuple(
        s_series_of_tuples: Series,
        i_index: int
) -> Series:

    # In case we have null values in the serie, we replace with empty tuple...
    #s_series_of_tuples.fillna(())

    _s_the_return: Series = s_series_of_tuples.apply(
        func=lambda i_rec: nan if (i_rec is None) else (i_rec[i_index] if (len(i_rec) >= i_index) else nan)
    )

    return _s_the_return



from inspect import stack
from logging import getLogger, Logger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestCheckConfigExcelDashboard(TestCase):

    _logger: Logger = getLogger(__name__)

    _the_config_checker: ConfigChecker = ConfigChecker(
        lst_widgets_json_schemas=[
            {
                "$id": "https://enovation.com/excel_dashboard/cover_sheet",
                "type": "object",
                "required": [
                    "narrative"
                ],
                "additionalProperties": False,
                "properties": {
                    "narrative": {
                        "type": "string"
                    }
                }
            }
        ]
    )

    def test_01_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self._the_config_checker.validate(
            dict_to_validate={
                "sheets": {
                    "test_sheet_name": {
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                "config": {
                                    "narrative": "the text to print for test"
                                },
                            }
                        }
                    }
                }
            },
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unexpected_node(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                ValidationError,
                f"is not valid under any of the given schemas"
        ):
            self._the_config_checker.validate(
                dict_to_validate={
                    "sheets": {
                        "test_sheet_namef": {
                            "widgets": {
                                "test_a_widget_label": {
                                    "address": "A1",
                                    "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                    "config": {
                                        "narrative": "the text to print for test",
                                        "test unexpected node": "jsg"
                                    },
                                }
                            }
                        }
                    }
                },
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_missing_node(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                ValidationError,
                f"is not valid under any of the given schemas"
        ):
            self._the_config_checker.validate(
                dict_to_validate={
                    "sheets": {
                        "test_sheet_namef": {
                            "widgets": {
                                "test_a_widget_label": {
                                    "address": "A1",
                                    "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                    "config": {
                                        # "narrative": "the text to print for test",
                                    },
                                }
                            }
                        }
                    }
                },
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import json
import os
import unittest
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from com_enovation.toolbox.data_broadcaster.data_broadcaster import DataBroadcaster
from com_enovation.toolbox.data_handler.data_handler import DataHandler


class TestFunction_InstantiateDataHandlers(unittest.TestCase):
    """
    Test the translation:
    - From the Data Broadcaster configuration, isolating the configurations for the Data Handlers
    - To the relevant Data Handlers, properly configured.
    """
    _logger: Logger = getLogger(__name__)

    def test_function_instantiate_data_handler(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the config
        with open(Path(os.path.join(os.path.dirname(__file__), 'IN_test_data/BroadcasterConfiguration.json')), 'r') \
                as json_file:
            dict_config: dict = json.load(json_file)

        # We load the expected result
        with open(Path(os.path.join(os.path.dirname(__file__), 'IN_test_data/ExpectedDataHandlerConfig.json')), 'r') \
                as json_file:
            dict_expected: dict = json.load(json_file)

        _dict_the_dhs: dict[str, DataHandler] = DataBroadcaster._instantiate_data_handlers(
            dict_distributions_configurations=dict_config[
                DataBroadcaster._const_str__config_tag__distributions_cfgs__as_dict
            ]
        )

        for k_dataset, v_dh in _dict_the_dhs.items():

            _dict_the_expected: dict = dict_expected[k_dataset]
            _dict_the_actual: dict = v_dh.dict_configuration

            self.assertDictEqual(
                d1=_dict_the_expected,
                d2=_dict_the_actual,
                msg="D1 (expected) is not equal to D2 (actual)."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack

from pandas import DataFrame, read_excel

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_AddEvalColumn(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_add_eval_column__vanilla(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__), '01.data_extract_for_data_quality_checks.xlsx'))
        )

        # We type the data extract...
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__type_columns(
            dict_columns={
                "StatusA": "str",
                "StatusB": "str"
            }
        )(df_data_extract)

        _lst_authorized: list[str] = [
            "OpenOpen",
            "OpenInactive",
            "InactiveOpen",
            "InactiveInactive",
            "ClosedClosed"
        ]

        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__add_eval_columns(
            str_expressions=f"""
                DQ_01=StatusA+StatusB=={[
                "OpenOpen",
                "OpenInactive",
                "InactiveOpen",
                "InactiveInactive",
                "ClosedClosed"
            ]}
                DQ_02=StatusA=='Open'
                """,
            b_copy=True
        )(df_data_extract)

        self.assertEqual(
            first=df_the_modified_dataframe.to_string(),
            second=f"    StatusA   StatusB  DQ_01  DQ_02\n"
                   f"0      Open      Open   True   True\n"
                   f"1      Open  Inactive   True   True\n"
                   f"2      Open    Closed  False   True\n"
                   f"3  Inactive      Open   True  False\n"
                   f"4  Inactive  Inactive   True  False\n"
                   f"5  Inactive    Closed  False  False\n"
                   f"6    Closed      Open  False  False\n"
                   f"7    Closed  Inactive  False  False\n"
                   f"8    Closed    Closed   True  False",
            msg=""
        )
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from copy import deepcopy
from inspect import stack

from logging import Logger, getLogger

from pandas import DataFrame

from com_enovation.toolbox.data_handler.data_checker.data_check_feature_factory import DataCheckFeatureFactory
from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory import DataModificationFeatureFactory


class DataHandler:
    """
    Capabilities to easily handle data through configuration (rather than coding):
    - Enrich: modify the data set
    - Check: check the data set without altering it.

    Configuration is of the kind:
        {
            "A label for the sequence that can be run in isolation": {
                "description": "A description of the sequence, useful to monitor and debug the execution",
                "steps": [
                    {
                        "description": "Description of the step 1 that is executed while calling the sequence",
                        "feature": "The name of the feature that is called",
                        "param": {
                            ... Depending on the feature, parameters are different ...,
                            "b_copy":false  # or true in case we need to copy the dataframe before it is modified
                                            # only valable for "modification" features, not relevant for "checks"
                        }
                    }
                ]
            }
        }
    """

    # Instance properties (decorated with @property)
    @property
    def dict_configuration(self) -> dict:
        return self._dict_configuration

    @property
    def dict_callable_sequence(self) -> dict:
        return self._dict_callable_sequence

    # Constants
    const_str__config_tag__description__as_str: str = "description"
    const_str__config_tag__steps__as_list: str = "steps"
    const_str__config_tag__feature__as_str: str = "feature"
    const_str__config_tag__param__as_str: str = "param"

    const_str__config_tag__callable__as_callable: str = "callable"
    const_str__config_tag__callable__type__as_str: str = "type"
    const_str__config_tag__callable__type__check__as_str: str = "check"
    const_str__config_tag__callable__type__modification__as_str: str = "modification"

    # Class properties
    _logger: Logger = getLogger(__name__)

    def __init__(
            self,
            dict_config: dict
    ):
        """
        Initialize a data handler instance using its configuration.
        Once initialized, sequences can be called using the function handle_data
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self._dict_configuration: dict = \
            self._check_dict_config(dict_config=dict_config)

        self._dict_callable_sequence: dict = \
            self._instantiate_callable_sequence(dict_config=dict_config)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def handle_data(
            self,
            df_data: DataFrame,
            sequence_alias: str | list[str] = None
    ) -> DataFrame:
        """
        Execute a sequence or a list of sequences.

        In case:
        - No sequence is provided, all the sequences are executed
        - A sequence to execute is not configured, an exception is raised.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if sequence_alias is None:
            _lst_the_sequence_alias: list[str] = [i_seq for i_seq in self.dict_configuration]
        elif isinstance(sequence_alias, str):
            _lst_the_sequence_alias: list[str] = [sequence_alias]
        else:
            _lst_the_sequence_alias: list[str] = sequence_alias

        # We set the return value as the input dataframe
        _df_the_return: DataFrame = df_data

        # We run each sequence one by one
        for i_sequence in _lst_the_sequence_alias:

            # We check the sequence we call is effectively configured
            if i_sequence not in self.dict_configuration:
                raise Exception(f"The sequence '{i_sequence}' is not defined in the configuration, which contains "
                                f"the following sequences: '{', '.join(self.dict_configuration)}'.")

            # We get the sequence configuration
            _dict_the_sequence_config: dict = self.dict_callable_sequence[i_sequence]

            self._logger.info(f"Sequence '{i_sequence}' is starting.")

            # We loop through each and every steps
            for i_dict_step_configuration in _dict_the_sequence_config[self.const_str__config_tag__steps__as_list]:

                _fn: callable = i_dict_step_configuration[self.const_str__config_tag__callable__as_callable]

                if i_dict_step_configuration[self.const_str__config_tag__callable__type__as_str] == \
                        self.const_str__config_tag__callable__type__modification__as_str:

                    _df_the_return = _fn(_df_the_return)
                    self._logger.info(
                        f"\tStep '{i_dict_step_configuration[self.const_str__config_tag__feature__as_str]} - "
                        f"{i_dict_step_configuration[self.const_str__config_tag__description__as_str]}' returned.")

                else:
                    _s_the_return = _fn(_df_the_return)

                    if _s_the_return is None:
                        self._logger.info(
                            f"\tStep '{i_dict_step_configuration[self.const_str__config_tag__feature__as_str]} - "
                            f"{i_dict_step_configuration[self.const_str__config_tag__description__as_str]}' returned "
                            f"None.")

                    else:
                        _s_the_unexpected_values: str = ""
                        for k, v in _s_the_return.items():
                            _s_the_unexpected_values += "\n\t- "+str(k)+": "+str(v)
                        raise Exception(
                            f"-----------------\n"
                            f"Step '{i_dict_step_configuration[self.const_str__config_tag__feature__as_str]} - "
                            f"{i_dict_step_configuration[self.const_str__config_tag__description__as_str]}' returned "
                            f"the following: {_s_the_unexpected_values}"
                            f"\n-----------------")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return _df_the_return

    def _check_dict_config(self, dict_config: dict) -> dict:
        """
        TODO: Implement the json schema to check the configuration
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")
        self._logger.warning(f"TODO: the configuration json schema is not yet implemented.")
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return dict_config

    def _instantiate_callable_sequence(self, dict_config: dict) -> dict:
        """
        We enrich the data handler configuration with callable instances of the features configured, that can then be
        easily called within "handle_data".
        We return a duplicated (deep copied) instance of dict_config, not to alter the original object.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We instantiate the feature factories
        _the_modification_features_factory: DataModificationFeatureFactory = DataModificationFeatureFactory()
        _the_check_features_factory: DataCheckFeatureFactory = DataCheckFeatureFactory()

        # We copy the sequence configuration, so we keep all the details in the callable sequence...
        _dict_the_return: dict = deepcopy(dict_config)

        # We then enrich sequence by sequence
        for i_sequence_label, i_sequence_config in _dict_the_return.items():

            # We then enrich step by step
            for i_step in i_sequence_config[self.const_str__config_tag__steps__as_list]:

                _modif_fn: callable = _the_modification_features_factory.get_modification_feature(
                    str_feature_label=i_step[self.const_str__config_tag__feature__as_str],
                    **i_step[self.const_str__config_tag__param__as_str]
                )

                _check_fn: callable = _the_check_features_factory.get_check_feature(
                    str_feature_label=i_step[self.const_str__config_tag__feature__as_str],
                    **i_step[self.const_str__config_tag__param__as_str]
                )

                if (_modif_fn is None) & (_check_fn is None):
                    raise Exception(
                        f"No feature labelled '{i_step[self.const_str__config_tag__feature__as_str]}' could be "
                        f"found. Be aware that only the following features exist:"
                        f"\nMODIFICATION FEATURES: "
                        f"{', '.join(_the_modification_features_factory.get_modifications_features_labels())}"
                        f"\nCHECK FEATURES: {', '.join(_the_check_features_factory.get_check_features_labels())}")

                if (_modif_fn is not None) & (_check_fn is not None):
                    self._logger.error(
                        f"Feature labelled '{i_step[self.const_str__config_tag__feature__as_str]}' is ambiguous "
                        f"as it is defined both as a CHECK and a MODIFICATION feature. It will arbitrarily be "
                        f"considered as a MODIFICATION feature. To fix this issue, you need to rename the CHECK "
                        f"feature '{i_step[self.const_str__config_tag__feature__as_str]}'.")

                if _modif_fn:
                    i_step[self.const_str__config_tag__callable__as_callable] = _modif_fn
                    i_step[self.const_str__config_tag__callable__type__as_str] = \
                        self.const_str__config_tag__callable__type__modification__as_str

                else:
                    i_step[self.const_str__config_tag__callable__as_callable] = _check_fn
                    i_step[self.const_str__config_tag__callable__type__as_str] = \
                        self.const_str__config_tag__callable__type__check__as_str

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return _dict_the_return



import logging
from inspect import stack

import click
from click import BadParameter
from com_enovation.helper.click.python_literal_argument_and_option import PythonLiteralArgument
from pandas import DataFrame

from com_enovation.toolbox.data_handler.data_handler import DataHandler

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dh-sequence')
@click.pass_context
@click.argument('alias-in-data-extract', type=str)
@click.argument('alias-in-config', type=str)
@click.argument('sequence', type=str)
@click.argument('alias-out-data-handled', type=str)
def dh_sequence(ctx_context, alias_in_data_extract, alias_in_config, sequence, alias_out_data_handled):
    """
    Handle a data frame
    - Load dataframe labelled ALIAS_IN_DATA_EXTRACT in context
    - Execute the SEQUENCE that is configured in a dictionary labelled ALIAS_IN_CONFIG in context
    - Persist the output into context as a dataframe labelled ALIAS_OUT_DATA_HANDLED.

    Note: this command only executes one sequence. In case you need to execute a list of sequences, use the equivalent
    command dh-sequences (with an 's')
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    # ##################################################################################################################
    # ##################################################################################################################
    # WE CHECK ARGUMENTS
    # ##################################################################################################################
    # ##################################################################################################################

    # ##################################################################################################################
    # 1. ALIAS IN CONFIG
    if alias_in_config not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot find argument '{alias_in_config}' in context, as the dictionary configuration with the "
                    f"logic to handle your data.",
            param_hint=f"Ensure you effectively load in the context a dictionary configuration with the logic to "
                       f"handle your data.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in_config], dict):
        raise BadParameter(
            message=f"Argument '{alias_in_config}' is of type '{type(ctx_context.obj[alias_in_config])}' while "
                    f"'dict' is expected.",
            param_hint=f"Ensure you effectively provide an argument '{alias_in_config}' of type 'dict'.",
            ctx=ctx_context
        )

    # ##################################################################################################################
    # 2. ALIAS IN DATA EXTRACT
    if alias_in_data_extract not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot find argument '{alias_in_data_extract}' in context, as the data extraction to handle.",
            param_hint=f"Ensure you effectively load in the context a dataframe with the data extraction to handle.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in_data_extract], DataFrame):
        raise BadParameter(
            message=f"Argument '{alias_in_data_extract}' is of type '{type(ctx_context.obj[alias_in_data_extract])}' "
                    f"while 'DataFrame' is expected.",
            param_hint=f"Ensure you effectively provide an argument '{alias_in_data_extract}' of type 'DataFrame'.",
            ctx=ctx_context
        )

    # ##################################################################################################################
    # 3. ALIAS OUT DATA HANDLED
    if alias_out_data_handled in ctx_context.obj:
        _logger.warning(f"Argument '{alias_out_data_handled}' exists in context, of type "
                        f"'{type(ctx_context[alias_out_data_handled])}'. It will be overriden.")

    # ##################################################################################################################
    # ##################################################################################################################
    # WE EXECUTE THE LOGIC
    # ##################################################################################################################
    # ##################################################################################################################

    try:
        _the_data_handler: DataHandler = DataHandler(
            dict_config=ctx_context.obj[alias_in_config]
        )
    except Exception as e_exception:
        _logger.fatal(f"An exception was raised when instantiating the DataHandler...", exc_info=e_exception)
        raise e_exception

    try:
        _df_the_enriched_data: DataFrame = _the_data_handler.handle_data(
            df_data=ctx_context.obj[alias_in_data_extract],
            sequence_alias=sequence
        )
    except Exception as e_exception:
        _logger.fatal(f"An exception was raised when running sequence '{sequence}...", exc_info=e_exception)
        raise e_exception

    ctx_context.obj[alias_out_data_handled] = _df_the_enriched_data
    ctx_context.obj["_" + alias_out_data_handled] = {"src": "dh-sequence"}

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('dh-sequences')
@click.pass_context
@click.argument('alias-in-data-extract', type=str)
@click.argument('alias-in-config', type=str)
@click.argument('sequences', cls=PythonLiteralArgument, type=list)
@click.argument('alias-out-data-handled', type=str)
def dh_sequences(ctx_context, alias_in_data_extract, alias_in_config, sequences, alias_out_data_handled):
    """
    Handle a data frame
    - Load dataframe labelled ALIAS_IN_DATA_EXTRACT in context
    - Execute each sequence from SEQUENCES that are configured in a dictionary labelled ALIAS_IN_CONFIG in context
    - Persist the output into context as a dataframe labelled ALIAS_OUT_DATA_HANDLED.

    Note: this command executes a list of sequences. In case you need to execute one single sequence, use the
    equivalent command dh-sequence (without any 's')
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    # ##################################################################################################################
    # ##################################################################################################################
    # WE CHECK ARGUMENTS...
    # ##################################################################################################################
    # ##################################################################################################################

    # ##################################################################################################################
    # 1. ALIAS IN CONFIG
    if alias_in_config not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot find argument '{alias_in_config}' in context, as the dictionary configuration with the "
                    f"logic to handle your data.",
            param_hint=f"Ensure you effectively load in the context a dictionary configuration with the logic to "
                       f"handle your data.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in_config], dict):
        raise BadParameter(
            message=f"Argument '{alias_in_config}' is of type '{type(ctx_context.obj[alias_in_config])}' while "
                    f"'dict' is expected.",
            param_hint=f"Ensure you effectively provide an argument '{alias_in_config}' of type 'dict'.",
            ctx=ctx_context
        )

    # ##################################################################################################################
    # 2. ALIAS IN DATA EXTRACT
    if alias_in_data_extract not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot find argument '{alias_in_data_extract}' in context, as the data extraction to handle.",
            param_hint=f"Ensure you effectively load in the context a dataframe with the data extraction to handle.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in_data_extract], DataFrame):
        raise BadParameter(
            message=f"Argument '{alias_in_data_extract}' is of type '{type(ctx_context.obj[alias_in_data_extract])}' "
                    f"while 'DataFrame' is expected.",
            param_hint=f"Ensure you effectively provide an argument '{alias_in_data_extract}' of type 'DataFrame'.",
            ctx=ctx_context
        )

    # ##################################################################################################################
    # 3. ALIAS OUT DATA HANDLED
    if alias_out_data_handled in ctx_context.obj:
        _logger.warning(f"Argument '{alias_out_data_handled}' exists in context, of type "
                        f"'{type(ctx_context[alias_out_data_handled])}'. It will be overriden.")

    # ##################################################################################################################
    # ##################################################################################################################
    # WE EXECUTE THE LOGIC
    # ##################################################################################################################
    # ##################################################################################################################

    try:
        _the_data_handler: DataHandler = DataHandler(
            dict_config=ctx_context.obj[alias_in_config]
        )
    except Exception as e_exception:
        _logger.fatal(f"An exception was raised when instantiating the DataHandler...", exc_info=e_exception)
        raise e_exception

    _df_the_return: DataFrame = ctx_context.obj[alias_in_data_extract]

    try:
        _df_the_enriched_data: DataFrame = _the_data_handler.handle_data(
            df_data=ctx_context.obj[alias_in_data_extract],
            sequence_alias=sequences
        )
    except Exception as e_exception:
        _logger.fatal(f"An exception was raised when running one among the following sequences "
                      f"'{', '.join(sequences)}...", exc_info=e_exception)
        raise e_exception

    ctx_context.obj[alias_out_data_handled] = _df_the_enriched_data
    ctx_context.obj["_" + alias_out_data_handled] = {"src": "dh-sequence"}

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
from inspect import stack
from logging import getLogger, Logger
from pathlib import Path
from unittest import TestCase

import numpy
import xlsxwriter
from pandas import DataFrame, to_datetime
from pandas.testing import assert_frame_equal

from com_enovation.helper import excel_loader
from com_enovation.toolbox.excel_dashboard.excel_dashboarder import ExcelDashboarder
from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker
from com_enovation.toolbox.excel_dashboard.widgets.table.widget_table import WidgetTable


class TestWidget(TestCase):

    _logger: Logger = getLogger(__name__)

    _table_config: dict = {

        "options": {

            # to turn on or off the autofilter in the header row. It is on by default
            "autofilter": False,

            # to turn on or off the header row in the table. It is on by default
            "header_row": True,

            # to create columns of alternating color in the table. It is on by default
            "banded_columns": False,

            # to create rows of alternating color in the table. It is on by default
            "banded_rows": True,

            # to highlight the first column of the table. The type of highlighting will depend on the style of the
            # table. It may be bold text or a different color. It is off by default
            "first_column": True,

            # to highlight the last column of the table. The type of highlighting will depend on the style of the
            # table. It may be bold text or a different color. It is off by default
            "last_column": True,

            # to set the style of the table. Standard Excel table format names should be used (with matching
            # capitalization)
            "style": "Table Style Medium 15",

            # to turn on the total row in the last row of a table. It is distinguished from the other rows by a
            # different formatting and also with dropdown SUBTOTAL functions
            "total_row": True,

            # default tables are named Table1, Table2, etc. The name parameter can be used to set the name of the
            # table
            "name": "jsg_table_name",

            # to set properties for columns within the table
            "columns": [
                {
                    # the column header
                    "header": "jsg data col 1",

                    # the format for the column header
                    "header_format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    },

                    # total captions are specified via the columns property and the total_string sub properties
                    "total_string": "jsg data col 1 total",

                    # total functions are specified via the columns property and the total_function sub properties
                    # total_function is a string among: average, count_nums, count, max, min, std_dev, sum, var
                    "total_function": "count",

                    # to set a calculated value for the total_function using the total_value sub property. Only
                    # necessary when creating workbooks for applications that cannot calculate the value of
                    # formulas automatically. This is similar to setting the value optional property in
                    # write_formula()
                    "total_value": 1,

                    # "the format for the column"
                    "format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    }
                },
                {
                    # the column header
                    "header": "jsg formula col 2",

                    # the format for the column header
                    "header_format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    },

                    # column formulas applied using the column formula property
                    "formula": "=today()",

                    # total captions are specified via the columns property and the total_string sub properties
                    "total_string": "jsg today total",

                    # total functions are specified via the columns property and the total_function sub properties
                    # total_function is a string among: average, count_nums, count, max, min, std_dev, sum, var
                    "total_function": "average",

                    # to set a calculated value for the total_function using the total_value sub property. Only
                    # necessary when creating workbooks for applications that cannot calculate the value of
                    # formulas automatically. This is similar to setting the value optional property in
                    # write_formula()
                    "total_value": 15,

                    # "the format for the column"
                    "format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    }
                },
                {
                    # the column header
                    "header": "jsg data col 3",

                    # "the format for the column"
                    "format": {
                       "font_name": "Times New Roman",
                       "font_size": 15
                    },
                },
                {
                    # the column header
                    "header": "jsg formula col 4",

                    # the format for the column header
                    "header_format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    },

                    # # total captions are specified via the columns property and the total_string sub properties
                    "total_string": "jsg today total",
                    #
                    # # total functions are specified via the columns property and the total_function sub properties
                    # # total_function is a string among: average, count_nums, count, max, min, std_dev, sum, var
                    "total_function": "count",
                    #
                    # # to set a calculated value for the total_function using the total_value sub property. Only
                    # # necessary when creating workbooks for applications that cannot calculate the value of
                    # # formulas automatically. This is similar to setting the value optional property in
                    # # write_formula()
                    "total_value": 15,
                    #
                    # # "the format for the column"
                    # "format": {
                    #    "font_name": "Times New Roman",
                    #    "font_size": 15
                    #    # "num_format": 'mmm d yyyy hh:mm AM/PM'
                    # },

                    # column formulas applied using the column formula property
                    # JSG as of 07-MAR-2023: "formula": "=today()+5000",
                    "formula": "=[[#This Row],[jsg formula col 2]]+5000",
                },
                {
                    # the column header
                    "header": "jsg today ter header",

                    # the format for the column header
                    "header_format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                    },

                    # "the format for the column"
                    "format": {
                       "font_name": "Times New Roman",
                       "font_size": 15,
                       "num_format": 'MM/DD/YY'
                    },

                    # column formulas applied using the column formula property
                    "formula": "=DATE(2022,8,17)",
                },
            ]
        }
    }

    def test_01_config_checker(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker: ConfigChecker = ConfigChecker(
            lst_widgets_json_schemas=[WidgetTable.get_dict_config_schema()],
            str_base_json_schema_id=WidgetTable.get_widget_id()
        )

        _the_config_checker.validate(
            dict_to_validate=self._table_config,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unitary(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b'],
                [30, 'c']
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        # We instantiate the widget
        _the_widget: WidgetTable = WidgetTable(
            str_address="B2",
            dict_parameters_map={"df_data": "jsg_data"},
            dict_config=self._table_config,
            dict_raw_parameters={"jsg_data": _df_data},
            str_sheet_name="abc",
            dict_default_format={}
        )

        # We print it into an excel spreadsheet
        _the_workbook = xlsxwriter.Workbook(Path(os.path.join(os.path.dirname(__file__), 'TC02.excel.xlsx')))
        _the_worksheet = _the_workbook.add_worksheet()

        # We print the widget
        _the_widget.write_to_excel(
            wb_workbook=_the_workbook,
            ws_worksheet=_the_worksheet
        )

        # We close the workbook
        _the_workbook.close()

        # #############################################################################################################
        # We now read the Excel spreadsheet, and assert the result is as expected
        # #############################################################################################################
        _bean: excel_loader.ExcelLoaderBean = excel_loader.ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC02.excel.xlsx')
        )

        _df_read_data: DataFrame = _bean.dict_tables["jsg_table_name"]

        # Before doing anything, we expect:
        # - As many columns as configured
        # - One extra line, corresponding to the total line
        self.assertEqual(
            first=len(
                self._table_config[
                    WidgetTable.const_str__config_label__str_options
                ][
                    WidgetTable.const_str__config_label__str_columns
                ]
            ),
            second=len(_df_read_data.columns),
            msg="We expect the same number of columns as configured."
        )
        self.assertEqual(
            first=len(_df_data.index),
            second=len(_df_read_data.index)-1,
            msg="We expect one extra line in the excel spreadsheet, corresponding to the line total."
        )

        # We remove the formula columns, and apply the same name as the original data frame
        _df_read_data = _df_read_data[[
            i_column[WidgetTable.const_str__config_label__str_header]
            for i_column in self._table_config[
                WidgetTable.const_str__config_label__str_options
            ][
                WidgetTable.const_str__config_label__str_columns
            ]
            if WidgetTable.const_str__config_label__str_formula not in i_column
        ]]
        self.assertEqual(
            first=len(_df_data.columns),
            second=len(_df_read_data.columns),
            msg="After having removed the formula columns, we expect the same number of columns."
        )
        _df_read_data.columns = _df_data.columns

        # We remove the last total row
        _df_read_data = _df_read_data[:-1]

        assert_frame_equal(
            left=_df_data,
            right=_df_read_data,
            obj=f"Reading at the excel did not return the expected data.",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_application(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b'],
                [30, 'c']
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC03.excel.xlsx')),
            dict_config={
                "sheets": {
                    "test_sheet_name": {
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "B2",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": self._table_config,
                                "parameters_map": {"df_data": "jsg_data"}
                            }
                        }
                    }
                }
            },
            **{"jsg_data": _df_data}
        )

        # #############################################################################################################
        # We now read the Excel spreadsheet, and assert the result is as expected
        # #############################################################################################################
        _bean: excel_loader.ExcelLoaderBean = excel_loader.ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC03.excel.xlsx')
        )

        _df_read_data: DataFrame = _bean.dict_tables["jsg_table_name"]

        # Before doing anything, we expect:
        # - As many columns as configured
        # - One extra line, corresponding to the total line
        self.assertEqual(
            first=len(
                self._table_config[
                    WidgetTable.const_str__config_label__str_options
                ][
                    WidgetTable.const_str__config_label__str_columns
                ]
            ),
            second=len(_df_read_data.columns),
            msg="We expect the same number of columns as configured."
        )
        self.assertEqual(
            first=len(_df_data.index),
            second=len(_df_read_data.index)-1,
            msg="We expect one extra line in the excel spreadsheet, corresponding to the line total."
        )

        # We remove the formula columns, and apply the same name as the original data frame
        _df_read_data = _df_read_data[[
            i_column[WidgetTable.const_str__config_label__str_header]
            for i_column in self._table_config[
                WidgetTable.const_str__config_label__str_options
            ][
                WidgetTable.const_str__config_label__str_columns
            ]
            if WidgetTable.const_str__config_label__str_formula not in i_column
        ]]
        self.assertEqual(
            first=len(_df_data.columns),
            second=len(_df_read_data.columns),
            msg="After having removed the formula columns, we expect the same number of columns."
        )
        _df_read_data.columns = _df_data.columns

        # We remove the last total row
        _df_read_data = _df_read_data[:-1]

        assert_frame_equal(
            left=_df_data,
            right=_df_read_data,
            obj=f"Reading at the excel did not return the expected data.",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returned")

    def test_04_wrong_nb_of_columns(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a', ""],
                [20, 'b', ""],
                [30, 'c', ""]
            ],
            columns=['jsg numbers col1', 'jsg string col2', 'jsg string col3']
        )

        # We instantiate the widget
        with self.assertRaisesRegex(
                Exception,
                f"We do not have a coherent number of columns"
                f"[\\s\\S]*df_data contains '3' records"
                f"[\\s\\S]*configuration contains in total '5' columns"
                f"[\\s\\S]*among which, '2' are non formula"
        ):
            _the_widget: WidgetTable = WidgetTable(
                str_address="B2",
                dict_parameters_map={"df_data": "jsg_data"},
                dict_config=self._table_config,
                dict_raw_parameters={"jsg_data": _df_data},
                str_sheet_name="abc",
                dict_default_format={}
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_05_duplicated_columns_labels(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b'],
                [30, 'c']
            ],
            columns=['jsg numbers col1', 'jsg numbers col1']
        )

        # We instantiate the widget
        with self.assertRaisesRegex(
                Exception,
                f"In the configuration of the widget Table, you mentioned several columns with the same labels... "
                f"Check columns labelled 'jsg data col 1'."
        ):
            _the_widget: WidgetTable = WidgetTable(
                str_address="B2",
                dict_parameters_map={"df_data": "jsg_data"},
                dict_config={
                    "options": {
                        "columns": [
                            {"header": "jsg data col 1"},
                            {"header": "jsg data col 1"},
                            {"header": "jsg today ter header", "formula": "=DATE(2022,8,17)"}
                        ]
                    }
                },
                dict_raw_parameters={"jsg_data": _df_data},
                str_sheet_name="abc",
                dict_default_format={}
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_06_nan_values(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b'],
                [30, numpy.nan]
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC06.excel.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ],
                                        "total_row": True
                                    }
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )

    def test_07_total_row(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b']
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        # #############################################################################################################
        # 1. Without any total row!

        # 1.1. We produce the Excel file
        ExcelDashboarder().excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC07.excel.With_out_Totals.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ],
                                        "total_row": False,
                                        "name": "jsg_table_name"
                                    }
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )

        # 1.2. We read the Excel spreadsheet, and assert the result is as expected
        _bean: excel_loader.ExcelLoaderBean = excel_loader.ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC07.excel.With_out_Totals.xlsx')
        )

        _df_read_data: DataFrame = _bean.dict_tables["jsg_table_name"]

        # Before doing anything, we expect 2 lines (as no total!)
        self.assertEqual(
            first=len(_df_data.index),
            second=len(_df_read_data.index),
            msg="Given there is no total row, we expect the same number of lines in the excel as in the dataframe."
        )

        # #############################################################################################################
        # 2. Without a total row!

        # 2.1. We produce the Excel file
        ExcelDashboarder().excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC07.excel.WithTotals.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ],
                                        "total_row": True,
                                        "name": "jsg_table_name"
                                    }
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )

        # 2.2. We read the Excel spreadsheet, and assert the result is as expected
        _bean: excel_loader.ExcelLoaderBean = excel_loader.ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC07.excel.WithTotals.xlsx')
        )

        _df_read_data: DataFrame = _bean.dict_tables["jsg_table_name"]

        # Before doing anything, we expect 2 lines (as no total!)
        self.assertEqual(
            first=len(_df_data.index),
            second=len(_df_read_data.index)-1,
            msg="Given there is a total row, we expect one extra "
                "line in the excel spreadsheet compared to the dataframe"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_08_conditional_format__manual_check(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _df_data: DataFrame = DataFrame(
            [
                [10, 'a'],
                [20, 'b']
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        # #############################################################################################################
        # 1. Without any total row!

        # 1.1. We produce the Excel file
        ExcelDashboarder().excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__),
                                                 'TC08.excel.CondFormatWith_out_Totals.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ],
                                        "total_row": False,
                                        "name": "jsg_table_name"
                                    },
                                    "conditional_formattings": [
                                      {
                                        "range": "@A:@A",
                                        "options": {
                                            "type": "cell", "criteria": "equal to", "value": 20,
                                            "format": {"bg_color": "red"}
                                        }
                                      }
                                    ],
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )

        # #############################################################################################################
        # 2. Without a total row!

        # 2.1. We produce the Excel file
        ExcelDashboarder().excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC08.excel.CondFormatWithTotals.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ],
                                        "total_row": True,
                                        "name": "jsg_table_name"
                                    },
                                    "conditional_formattings": [
                                      {
                                        "range": "@A:@A",
                                        "options": {
                                            "type": "cell", "criteria": "equal to", "value": 20,
                                            "format": {"bg_color": "red"}
                                        }
                                      }
                                    ],
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )

        # MANUAL CHECK: conditional formatting should span across the whole table

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_09_nat_values(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _df_data: DataFrame = DataFrame(
            [
                [10, '2023-03-25'],
                [20, None],
            ],
            columns=['jsg numbers col1', 'jsg string col2']
        )

        _df_data['jsg string col2'] = to_datetime(
            _df_data['jsg string col2']
        ).dt.normalize()

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC09.excel.NatValues.xlsx')),
            dict_config={
                "workbook.options": {"nan_inf_to_errors": True},
                "sheets": {
                    "test_sheet_name": {
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/table",
                                "config": {
                                    "options": {
                                        "columns": [
                                            {"header": "jsg numbers col1"},
                                            {"header": "jsg string col2"}
                                        ]
                                    }
                                },
                                "parameters_map": {
                                    "df_data": "df_data"
                                }
                            }
                        }
                    }
                }
            },
            df_data=_df_data
        )



import os
from inspect import stack
from logging import getLogger, Logger
from unittest import TestCase
import click
from click.testing import CliRunner
from pandas import DataFrame

from com_enovation.enov import enov


class TestClickDhDataHandler(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        runner = CliRunner()

        enov.add_command(check)

        # noinspection PyTypeChecker
        result = runner.invoke(
            cli=enov,
            args=[
                # To get all the logs...
                '--verbose',

                # We load the data handler configuration
                'dict-load-json',
                os.path.join(os.path.dirname(__file__), '../01.config.json'),
                'enov_config',

                # We load the excel file
                'df-load-xls',
                os.path.join(os.path.dirname(__file__), '../02.data_to_handle.xlsx'),
                'enov_data',

                'dh-sequence',
                'enov_data',
                'enov_config',
                'first sequence',
                'enov_first_output',

                'dh-sequences',
                'enov_data',
                'enov_config',
                '[\"first sequence\", \"second sequence\"]',
                'enov_second_output',

                'check'
            ]
        )

        self.assertEqual(
            first=0,
            second=result.exit_code,
            msg=str(result.exception)
        )


@click.command('check')
@click.pass_context
def check(ctx_context):
    # We get the outputs from both tests
    _df_the_first_output: DataFrame = ctx_context.obj["enov_first_output"]
    _df_the_second_output: DataFrame = ctx_context.obj["enov_second_output"]

    # We check the number of rows and columns
    TestCase().assertEqual(
        first=len(_df_the_first_output.index), second=34,
        msg="First Output: wrong number of rows"
    )

    # We check the number of rows and columns
    TestCase().assertEqual(
        first=len(_df_the_first_output.columns), second=18,
        msg="First Output: wrong number of columns"
    )

    # We check the number of rows and columns
    TestCase().assertEqual(
        first=len(_df_the_second_output.index), second=34,
        msg="Second Output: wrong number of rows"
    )

    # We check the number of rows and columns
    TestCase().assertEqual(
        first=len(_df_the_second_output.columns), second=19,
        msg="Second Output: wrong number of columns"
    )



from datetime import datetime
from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame
from scipy.special import psi

from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com_enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class TestFunctionComputeSpotPredictabilityForAGivenKey(TestCase):
    """
    We test all the combinations when calling the function _compute_spot_predictability_for_a_given_key_no_resample.

    Each test function executes 3 test cases:
    - For given start and end date that will be the same across 3 test cases
    - We first compute predictability with a measure (22-May) > last measure (19-May)
    - We then compute predictability with a measure spot§on (aka 19-May), expected to be 1
    - We eventually conpute predictability with a measure (16-May) < last measure (19-May).

    In addition to call function _compute_spot_predictability_for_a_given_key_no_resample, we call the psi special
    function. This can be useful to debug in case of any issue.
    """

    _logger: Logger = getLogger(__name__)

    def function_to_test(
            self,
            s_test_case: str,
            dt_start: datetime,
            dt_end: datetime,
            f_aggregated_predictability: float,
            i_a_west: int = None,
            i_b_west: int = None,
            i_m_west: int = None,
            f_psi_result_west: float = None,
            i_a_east: int = None,
            i_b_east: int = None,
            i_m_east: int = None,
            f_psi_result_east: float = None
    ):
        """
        Helper function to easily execute the various combinations:
        - The effective call to the function compute spot predictability, in north, south and spot-on (aka when
          measure = last measure)
        - And the calls to the psi function, both in the western and eastern cones.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We instantiate the dataframes
        the_df_dataset_north: DataFrame = DataFrame(
            columns=[
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
                "date_end",
                DatePredictabilityComputer.str__output__column_label__duration
            ],
            data=[
                [dt_start, datetime(2021, 5, 22), dt_end, (dt_end-dt_start).days],
                [dt_end,  datetime(2021, 5, 19), dt_end, 0],
            ]
        )
        the_df_dataset_spot_on: DataFrame = DataFrame(
            columns=[
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
                "date_end",
                DatePredictabilityComputer.str__output__column_label__duration
            ],
            data=[
                [dt_start, datetime(2021, 5, 19), dt_end, (dt_end-dt_start).days],
                [dt_end,  datetime(2021, 5, 19), dt_end, 0],
            ]
        )
        the_df_dataset_south: DataFrame = DataFrame(
            columns=[
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
                "date_end",
                DatePredictabilityComputer.str__output__column_label__duration
            ],
            data=[
                [dt_start, datetime(2021, 5, 16), dt_end, (dt_end-dt_start).days],
                [dt_end,  datetime(2021, 5, 19), dt_end, 0],
            ]
        )

        # We call the function 3 times
        f_the_computed_aggregated_predictability_north: float = \
            DatePredictabilityComputer()._compute_date_predictability_for_a_given_key_not_resampled(
                df_measures=the_df_dataset_north
            )
        f_the_computed_aggregated_predictability_spot_on: float = \
            DatePredictabilityComputer()._compute_date_predictability_for_a_given_key_not_resampled(
                df_measures=the_df_dataset_spot_on
            )
        f_the_computed_aggregated_predictability_south: float = \
            DatePredictabilityComputer()._compute_date_predictability_for_a_given_key_not_resampled(
                df_measures=the_df_dataset_south
            )

        # We test the psi function
        if f_psi_result_west:
            self.assertAlmostEqual(
                f_psi_result_west,
                (i_b_west-i_a_west)*(psi(i_b_west)-psi(i_b_west+i_m_west))+i_m_west,
                msg=f"Test Case '{s_test_case}': the psi function does not return expected results for parameters:"
                    f"\n\t-i_a_west='{i_a_west}'"
                    f"\n\t-i_b_west='{i_b_west}'"
                    f"\n\t-i_m_west='{i_m_west}'",
            )
        if f_psi_result_east:
            self.assertAlmostEqual(
                f_psi_result_east,
                (i_b_east-i_a_east)*(psi(i_b_east)-psi(i_b_east+i_m_east))+i_m_east,
                msg=f"Test Case '{s_test_case}': the psi function does not return expected results for parameters:"
                    f"\n\t-i_a_east='{i_a_east}'"
                    f"\n\t-i_b_east='{i_b_east}'"
                    f"\n\t-i_m_east='{i_m_east}'",
            )

        # We test the function results
        self.assertAlmostEqual(
            f_aggregated_predictability,
            f_the_computed_aggregated_predictability_north,
            msg=f"Test Case '{s_test_case}': the function does not return expected results in North (measure > last "
                f"measure)."
        )

        # We test the function results
        self.assertAlmostEqual(
            f_aggregated_predictability,
            f_the_computed_aggregated_predictability_south,
            msg=f"Test Case '{s_test_case}': the function does not return expected results in South (measure < last"
                f"measure"
        )

        # We test the function results
        self.assertAlmostEqual(
            1,
            f_the_computed_aggregated_predictability_spot_on,
            msg=f"Test Case '{s_test_case}': the function does not return expected results when measure is spot on."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_01_WxW_WxW(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="01. WNW to WNW, WSW to WSW, W to W",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 4, 28),
            f_aggregated_predictability=0.8666007905138338,
            i_a_west=19,
            i_b_west=22,
            i_m_west=2,
            f_psi_result_west=1.7332015810276675,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_WxW_to_xW(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="02. WNW to NW (decreasing cone), WSW to SW (increasing cone), W to W",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 16),
            f_aggregated_predictability=0.7148562733369739,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_WxW_to_xxW(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="03. WNW to NNW, WSW to SSW, W to W",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 18),
            f_aggregated_predictability=0.649869339397249,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_04_WxW_to_last_measure(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="04. WNW to last measure, WSW to last measure, W to last measure",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 19),
            f_aggregated_predictability=0.6216141507278035,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_05_WxW_to_xxE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="05. WNW to NNE, WSW to SSE, W to E",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 20),
            f_aggregated_predictability=0.595713561114145,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_06_WxW_to_xE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="06. WNW to NE (increasing cone), WSW to SE (decreasing cone), W to E",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 22),
            f_aggregated_predictability=0.5498894410284415,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_07_WxW_to_ExE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="07. WNW to ENE, WSW to ESE, W to E",
            dt_start=datetime(2021, 4, 26),
            dt_end=datetime(2021, 5, 25),
            f_aggregated_predictability=0.5154181195427406,
            i_a_west=1,
            i_b_west=4,
            i_m_west=20,
            f_psi_result_west=14.297125466739478,
            i_a_east=1,
            i_b_east=4,
            i_m_east=2,
            f_psi_result_east=0.6499999999999995,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_08_xW_to_ExE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="08. NW (decreasing cone) to ENE, SW (increasing cone) to ESE, W to E",
            dt_start=datetime(2021, 5, 16),
            dt_end=datetime(2021, 5, 25),
            f_aggregated_predictability=0.6499999999999995/9,
            i_a_east=1,
            i_b_east=4,
            i_m_east=2,
            f_psi_result_east=0.6499999999999995,
        )

    def test_09_xxE_to_ExE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="09. NNE to ENE, SSE to ESE, E to E",
            dt_start=datetime(2021, 5, 20),
            dt_end=datetime(2021, 5, 25),
            f_aggregated_predictability=0.6499999999999995/5,
            i_a_east=1,
            i_b_east=4,
            i_m_east=2,
            f_psi_result_east=0.6499999999999995,
        )

    def test_10_xE_to_ExE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="10. NE (increasing cone) to ENE, SE (decreasing cone) to ESE, E to E",
            dt_start=datetime(2021, 5, 22),
            dt_end=datetime(2021, 5, 25),
            f_aggregated_predictability=0.6499999999999995/3,
            i_a_east=1,
            i_b_east=4,
            i_m_east=2,
            f_psi_result_east=0.6499999999999995,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_11_ExE_to_ExE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="11. ENE to ENE, ESE to ESE, E to E",
            dt_start=datetime(2021, 5, 23),
            dt_end=datetime(2021, 5, 25),
            f_aggregated_predictability=0.32499999999999973,
            i_a_east=1,
            i_b_east=4,
            i_m_east=2,
            f_psi_result_east=0.6499999999999995,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_12_xW_to_xxW(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="12. NW (decreasing cone) to NNW, SW (increasing cone) to SSW, W to W",
            dt_start=datetime(2021, 5, 16),
            dt_end=datetime(2021, 5, 18),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_13_xW_to_last_measure_north(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="13. NW (decreasing cone) to last measure, SW (increasing cone) to last measure, W to last "
                        "measure",
            dt_start=datetime(2021, 5, 16),
            dt_end=datetime(2021, 5, 19),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_14_xW_to_xxE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="14. NW (decreasing cone) to NNE, SW (increasing cone) to SSE, W to E",
            dt_start=datetime(2021, 5, 16),
            dt_end=datetime(2021, 5, 21),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_15_xW_to_xE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="15. NW (decreasing cone) to NE (increasing cone), SW (decreasing cone) to SE (increasing "
                        "cone), W to E",
            dt_start=datetime(2021, 5, 16),
            dt_end=datetime(2021, 5, 22),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_16_xxW_to_xxW(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="16. NNW to NNW, SSW to SSW, W to W",
            dt_start=datetime(2021, 5, 17),
            dt_end=datetime(2021, 5, 18),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_17_xxW_to_last_measure(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="17. NNW to last measure, SSW to last measure, W to last measure",
            dt_start=datetime(2021, 5, 17),
            dt_end=datetime(2021, 5, 19),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_18_xxW_to_xxE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="18. NNW to NNE, SSW to SSE, W to E",
            dt_start=datetime(2021, 5, 17),
            dt_end=datetime(2021, 5, 20),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_19_xxW_to_xE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="19. NNW to NE (increasing cone), SSW to SE (decreasing cone), W to E",
            dt_start=datetime(2021, 5, 17),
            dt_end=datetime(2021, 5, 22),
            f_aggregated_predictability=0.0,
        )

    def test_20_last_measure_to_xxE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="20. Last measure to NNE, last measure to SSE, last measure to E",
            dt_start=datetime(2021, 5, 19),
            dt_end=datetime(2021, 5, 21),
            f_aggregated_predictability=0.0,
        )

    def test_21_last_measure_to_xE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="21. Last measure to NE (increasing cone), last measure to SE (decreasing cone), last measure "
                        "to E",
            dt_start=datetime(2021, 5, 19),
            dt_end=datetime(2021, 5, 22),
            f_aggregated_predictability=0.0,
        )

    def test_22_xxE_to_xxE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="22. NNE to NNE, SSE to SSE, E to E",
            dt_start=datetime(2021, 5, 20),
            dt_end=datetime(2021, 5, 21),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_23_xxE_to_xE(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self.function_to_test(
            s_test_case="23. NNE to NE (increasing cone), SSE to SE (decreasing cone), E to E",
            dt_start=datetime(2021, 5, 20),
            dt_end=datetime(2021, 5, 22),
            f_aggregated_predictability=0.0,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    # Je suis ici.
    # Je dois faire les tests unitaires lorsque je suis spot on
    # Et les tests au sud.

    def todo_test_full(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_df_dataset: DataFrame = DataFrame(
            columns=[
                DatePredictabilityComputer.str__input__column_label__key,
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
                "date_end",
                DatePredictabilityComputer.str__output__column_label__duration
            ],
            data=[
                ['abc', '2021-4-26', '2021-5-22', '2021-4-27', 1],
                ['abc', '2021-4-27', '2021-6-5', '2021-4-30', 3],
                ['abc', '2021-4-30', '2021-6-3', '2021-5-1', 1],
                ['abc', '2021-5-1', '2021-6-4', '2021-5-6', 5],
                ['abc', '2021-5-6', '2021-5-29', '2021-5-7', 1],
                ['abc', '2021-5-7', '2021-5-25', '2021-5-9', 2],
                ['abc', '2021-5-9', '2021-5-21', '2021-5-11', 2],
                ['abc', '2021-5-11', '2021-5-19', '2021-5-13', 2],
                ['abc', '2021-5-13', '2021-5-16', '2021-5-16', 3],
                ['abc', '2021-5-16', '2021-5-17', '2021-5-18', 2],
                ['abc', '2021-5-18', '2021-5-21', '2021-5-23', 5],
                ['abc', '2021-5-23', '2021-5-14', '2021-5-25', 2],
                ['abc', '2021-5-25', '2021-5-27', '2021-5-29', 4],
                ['abc', '2021-5-29', '2021-5-20', '2021-6-6', 8],
                ['abc', '2021-6-6', '2021-5-19', '2021-6-6', 0],
            ]
        )

        # We type the columns
        the_df_dataset = PandasDataframeTyper.type(
            df_to_type=the_df_dataset,
            dict_columns_to_type={
                DatePredictabilityComputer.str__input__column_label__date: PandasDataframeTyper.str__type__date,
                DatePredictabilityComputer.str__input__column_label__measure: PandasDataframeTyper.str__type__date,
                "date_end": PandasDataframeTyper.str__type__date,
            },
            b_strict=False
        )

        f_the_aggregated_predictability: float = \
            DatePredictabilityComputer()._compute_date_predictability_for_a_given_key_not_resampled(
                df_measures=the_df_dataset[[
                    DatePredictabilityComputer.str__input__column_label__key,
                    DatePredictabilityComputer.str__input__column_label__date,
                    DatePredictabilityComputer.str__input__column_label__measure,
                    "date_end",
                    DatePredictabilityComputer.str__output__column_label__duration
                ]]
            )

        self.assertEqual(
            0.4023347872297693,
            f_the_aggregated_predictability,
            msg=f"The computation did not return expected results",
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
